from ..constants import (
    BASE_URL,
)
from .models import CustomerMetadata, Customer
from ..base.client import BaseClient
from logging import getLogger
from ..utils.helpers import prepare_data
from typing import Optional

logger = getLogger(__name__)


class CustomerClient(BaseClient):
    def create(self, customer: Optional[CustomerMetadata | dict] = None, **kwargs) -> Customer:
        """creates a new customer using an or named arguments

        Args:
            customer (Optional[CustomerMetadata  |  dict], optional): You customer data, it can be \
            a dict, an instance of `abacatepay.customers.CustomerMetadata`.

        Returns:
            Customer: An instance of the new customer.
        """
        logger.debug(f"Creating customer with URL: {BASE_URL}/customer/create")

        json_data = prepare_data(customer or kwargs, CustomerMetadata)
        response = self._request(
            f"{BASE_URL}/customer/create",
            method="POST",
            json=json_data,
        )
        return Customer(**response.json()["data"])

    def list(self) -> list[Customer]:
        logger.debug(f"Listing customers with URL: {BASE_URL}/customer/list")
        response = self._request(f"{BASE_URL}/customer/list", method="GET")
        return [Customer(**bill) for bill in response.json()["data"]]



from pydantic import BaseModel


def prepare_data(data, model_cls: type[BaseModel], *, use_alias: bool = True) -> dict:
    """It formats the data to be in compliance with the given model class to be sent to the API

    Args:
        data (BaseModel | dict): the user input data.
        model_cls (type[BaseModel]): the model that will dumps.
        use_alias (bool, optional): if `True` dumps by alias. Defaults to True.
    """
    if isinstance(data, model_cls):
        return data.model_dump(by_alias=use_alias)
    
    elif isinstance(data, dict):
        return model_cls.model_validate(data).model_dump(by_alias=use_alias)
    
    raise TypeError('Invalid data type.')




from typing import Annotated, Iterator, List, Optional

from pydantic import (
    AliasChoices,
    AwareDatetime,
    BaseModel,
    Field,
    StringConstraints,
)

from ..constants import BILLING_KINDS, BILLING_METHODS, BILLING_STATUS
from ..customers.models import CustomerID, CustomerInline, CustomerMetadata
from ..products.models import Product, ProductInline

HttpUrl = Annotated[
    str,
    StringConstraints(
        pattern=(
            r"^(https?:\/\/)"  # Protocol (http or https)
            r"(([a-zA-Z0-9-]+\.)+[a-zA-Z]{2,})"  # Domain (including subdomains)
            r"(\/[^\s]*)?$" # Optional path
        )
    )
]

class BillingIn(BaseModel):
    """Represents a billing model to creation

    Args:
        frequency (BILLING_KINDS): The billing frequency. Defaults to `ONE_TIME`.
        methods (List[BILLING_METHODS]): the allowed method to the billing. Defaults to ['PIX']
        products (List[Product]): The list of products in the billing.
        completion_url (HttpUrl): the redirect URL when the payment is completed.
        return_url (HttpUrl): the redirect URL when the user clicks on the "back" button.
        customer_id (Optional[CustomerID]): Unique identifier of the billing.
        customer (CustomerMetadata): Your customer information. Defaults to {}.
    """
    frequency: BILLING_KINDS = 'ONE_TIME'
    methods: List[BILLING_METHODS] = ['PIX']
    products: List[Product]
    return_url: HttpUrl = Field(serialization_alias='returnUrl')
    completion_url: HttpUrl = Field(serialization_alias='completionUrl')
    customer_id: Optional[CustomerID] = Field(None, serialization_alias='customerId')
    customer: CustomerMetadata | dict = {}


class BillingMetadata(BaseModel):
    """model of a billing metadata

    Args:
        completion_url (str): the redirect URL when the payment is completed.
        return_url (str): the redirect URL when the user clicks on the "back" button.
        fee (int): the billing fee.
    """
    completion_url: str = Field(
        validation_alias=AliasChoices('completion_url', 'completionUrl'),
        serialization_alias='completionUrl',
    )
    return_url: str = Field(
        validation_alias=AliasChoices('return_url', 'returnUrl'),
        serialization_alias='returnUrl',
    )
    fee: int


class Billing(BaseModel):
    """Billing model

    Args:
        id (str): Unique identifier of the billing.
        url (str): the URL which the user can complete the payment.
        amount (int): the amount to be paid (in cents).
        status (BILLING_STATUS): the current billing status.
        dev_mode (bool): if it's operating in dev mode.
        methods (List[BILLING_METHODS]): the allowed method to the billing. 
        products (List[ProductInline]): The list of products in the billing.
        frequency (BILLING_KINDS): The billing frequency.
        next_billing (Optional[AwareDatetime]): date and time of the next billing.
        customer (Optional[CustomerInline]): the data of the customer that the billing belongs to.
        created_at (AwareDatetime): The date and time when the billing was created.
        updated_at (AwareDatetime): the date and time of the last billing update.
        allow_cupons (bool): If the billing has allowed coupons or not.
        coupons (List[Coupon]): the available coupons.
        coupons_used (List[Coupon]): active coupons for this billing.
        metadata (BillingMetadata): the billing metadata.
    """
    id: str
    url: str
    amount: int
    status: BILLING_STATUS
    dev_mode: bool = Field(alias='devMode')
    methods: List[BILLING_METHODS]
    products: List[ProductInline]
    frequency: BILLING_KINDS
    next_billing: Optional[AwareDatetime] = Field(
        None, 
        validation_alias=AliasChoices('next_billing', 'nextBilling'),
        serialization_alias='nextBilling',
    )
    customer: Optional[CustomerInline] = None
    created_at: AwareDatetime = Field(alias='createdAt')
    updated_at: AwareDatetime = Field(alias='updatedAt')
    allow_coupons: bool = Field(
        validation_alias=AliasChoices('allowCoupons', 'allow_coupons')
    )
    coupons: List[str]
    coupons_used: List[str] = Field(alias='couponsUsed')
    metadata: BillingMetadata


class BillingList(BaseModel):
    """Represents a list of Billings. It can be directly iterated and 
    supports the usage of the `len` function.
    
    Args:
        data (List[Billing]): the list object having all the billings.
    """
    data: List[Billing]

    def __len__(self) -> int:
        return len(self.data)
    
    def __iter__(self) -> Iterator[Billing]:
        return self.data.__iter__()



from .client import BillingClient
from .models import (
    Billing,
    BillingList,
    BillingIn,
    BillingMetadata,
)

__all__ = [
    "BillingClient",
    "Billing",
    "BillingList",
    "BillingIn",
    "BillingMetadata",
]



import requests
from typing import Literal
from ..constants import USER_AGENT
from ..utils.exceptions import raise_for_status, APITimeoutError, APIConnectionError


class BaseClient:
    def __init__(self, api_key: str):
        self.__api_key = api_key

    def _request(
        self,
        url: str,
        method: Literal["GET", "POST", "PUT", "PATCH", "DELETE"] = "GET",
        **kwargs,
    ):
        request = requests.Request(
            method,
            url,
            headers={
                "Authorization": f"Bearer {self.__api_key}",
                "User-Agent": USER_AGENT,
            },
            **kwargs,
        )
        try:
            prepared_request = request.prepare()
            with requests.Session() as s:
                response = s.send(prepared_request)

            raise_for_status(response)
            return response

        except requests.exceptions.Timeout:
            raise APITimeoutError(request=request)

        except requests.exceptions.ConnectionError:
            raise APIConnectionError(message="Connection error.", request=request)



import requests
from typing import Literal


class APIError(Exception):
    """The exception was raised due to an API error."""

    message: str
    request: str

    def __init__(self, message: str, request: requests.Request) -> None:
        super().__init__(message)
        self.message = """The exception was raised due to an API error."""
        self.request = request

    def __str__(self):
        if self.message:
            return f"{self.message}"


class APIStatusError(APIError):
    """Raised when an API response has a status code of 4xx or 5xx."""

    response: requests.Response
    status_code: int

    def __init__(
        self, message: str | None = None, *, response: requests.Response
    ) -> None:
        super().__init__(message, response.request)
        self.response = response
        self.status_code = response.status_code


class ForbiddenRequest(APIStatusError):
    """
    Means that the request was unsuccessful due to a forbidden request. Maybe your API key is wrong?
    """

    message: str | None
    status_code: Literal[401] = 401

    def __init__(self, response: requests.Response, message: str | None = None):
        super().__init__(message, response=response)
        self.message = "Means that the request was unsuccessful due to a forbidden request. Maybe your API key is wrong?"
        self.response = response
        self.status_code: Literal[401] = 401

    def __str__(self):
        if self.message:
            return f"{self.message} \n Status Code: {self.response.status_code} | Response: {self.response.text}"
        return self.response.content


class UnauthorizedRequest(APIStatusError):
    """
    Means that the request was unsuccessful due to a forbidden request. Maybe your API key doesn't have enought permissions
    """

    message: str | None
    status_code: Literal[403] = 403

    def __init__(self, response: requests.Response, message: str | None = None):
        super().__init__(message, response=response)
        self.message = "Means that the request was unsuccessful due to a forbidden request. Maybe your API key doesn't have enought permissions"
        self.response = response
        self.status_code: Literal[403] = 403

    def __str__(self):
        if self.message:
            return f"{self.message} | Status Code: {self.response.status_code} | Response: {self.response.text}"
        return self.response.content


class APIConnectionError(APIError):
    """The request was unsuccessful due to a connection error. Check your internet connection"""

    def __init__(
        self,
        *,
        message: str = "The request was unsuccessful due to a connection error. Check your internet connection",
        request: requests.Request,
    ) -> None:
        super().__init__(message, request)


class APITimeoutError(APIConnectionError):
    """The request got timed out. You might try checking your internet connection."""

    def __init__(self, request: requests.Request) -> None:
        super().__init__(
            message="Request timed out. Check your internet connection", request=request
        )


class BadRequestError(APIStatusError):
    """The request was unsuccessful due to a bad request. Maybe the request syntax is wrong"""

    status_code: Literal[400] = 400

    def __init__(self, response: requests.Response) -> None:
        self.response = response

    def __str__(self):
        return f"The request was unsuccessful due to a bad request. Maybe the request syntax is wrong. Message error: {self.response.json()}"


class NotFoundError(APIStatusError):
    status_code: Literal[404] = 404

    def __str__(self):
        return """The request was unsuccessful due to a not found error. Error status 404 | Requested URL: {}""".format(
            self.response.url
        )


class InternalServerError(APIStatusError):
    """The request was unsuccessful due to an internal server error."""

    status_code: Literal[500] = 500

    def __init__(self, response: requests.Response) -> None:
        super().__init__(
            message="The request was unsuccessful due to an internal server error. It's not your fault, just try again later.",
            response=response,
        )


def raise_for_status(response: requests.Response) -> None:
    code_exc_dict = {
        400: BadRequestError(response=response),
        401: UnauthorizedRequest(response=response),
        403: ForbiddenRequest(response=response),
        404: NotFoundError(response=response),
        500: InternalServerError(response=response),
    }

    code = response.status_code
    if code == 200:
        return
    
    if code not in code_exc_dict  and code >= 400:
        raise APIStatusError(message=response.text, response=response)
    
    raise code_exc_dict.get(
        response.status_code,
        APIError(message=response.text, request=response.request)
    )
    



from typing import Literal

MINIMUM_VALUE = 100
BASE_URL = "https://api.abacatepay.com/v1"
VERSION = "1.0.0"
USER_AGENT = f"Python SDK {VERSION}"

BILLING_STATUS = Literal['PENDING', 'EXPIRED', 'CANCELLED', 'PAID', 'REFUNDED']
BILLING_METHODS = Literal['PIX']
BILLING_KINDS = Literal['ONE_TIME', 'MULTIPLE_PAYMENTS']
DISCOUNT_KINDS = Literal['PERCENTAGE', 'FIXED']
COUPON_STATUS = Literal['PENDING', 'EXPIRED', 'CANCELLED', 'PAID', 'REFUNDED']



from logging import getLogger
from typing import Optional
from ..base.client import BaseClient
from ..constants import BASE_URL
from .models import Billing, BillingIn, BillingList
from ..utils.helpers import prepare_data

logger = getLogger(__name__)


class BillingClient(BaseClient):
    def create(self, data: Optional[BillingIn | dict] = None, **kwargs) -> Billing:
        """
        Create a new billing.

        Args:
            data (BillingIn): an instance of `abacatepay.billings.models.BillingIn` a dict \
            or the named params following the model schema.
        
        Keyword args:
            products (List[Product]): List of products to be billed.
            returnURL (str): The URL the user will be redirected to after the billing is completed.
            completionUrl (str): The URL the API will make a POST request after the billing is completed.
            methods (List[BILLING_METHODS]): The payment methods to be accepted. Defaults to ["PIX"].
            frequency (BILLING_KINDS): The frequency of the billing. Defaults to "ONE_TIME".
            customerId (str): The ID of the customer. If provided, the customer information won't be required.
            customer (CustomerMetadata): The customer information. If customerId is provided, this parameter is ignored.

        Returns:
            Billing: The response with the billing data.
        """
        json_data = prepare_data(data or kwargs, BillingIn)
        logger.debug('creating billing: %s', json_data)

        response = self._request(
            f"{BASE_URL}/billing/create",
            method="POST",
            json=json_data,
        )
        return Billing(**response.json()["data"])

    def list(self) -> BillingList:
        """
        List all bills.

        Returns:
            BillingList: A list of billing objects.
        """
        logger.debug(f"Listing bills with URL: {BASE_URL}/billing/list")
        response = self._request(f"{BASE_URL}/billing/list", method="GET")
        return BillingList.model_validate({'data': response.json()["data"]})



"""
The Python SDK for the AbacatePay API

Basic usage:
```python
import abacatepay
from abacatepay.models import Product

token = "<your api token>"
client = AbacatePay(token)

billing = client.billing.create(products=[Product(externalId="123", name="Teste", quantity=1, price=101, description="Teste")], returnURL="https://abacatepay.com", completionUrl="https://abacatepay.com")
print(billing.data.url)
# > https://abacatepay.com/pay/aaaaaaa
```

More examples found on https://abacatepay.readme.io/
"""

from .billings import BillingClient
from .customers import CustomerClient

class AbacatePay:
    def __init__(self, api_key: str):
        self.billing = BillingClient(api_key)
        self.customers = CustomerClient(api_key)
